@tool

extends Node3D

@export var tracking : UltraleapDeviceNode
@export var handedness : UltraleapTypes.Chirality = UltraleapTypes.Chirality.Left
@export var skeleton : Skeleton3D
@export var ex_map : RigMap
@export var rig_settings : RigSettings
@export var use_interpolation : bool = true


func _ready():
	if skeleton == null:
		return

	skeleton.animate_physical_bones = true


func _process(_delta):
	if tracking == null:
		skeleton.hide()
		return
		
	if Engine.is_editor_hint() and not tracking.run_in_editor:
		skeleton.reset_bone_poses()
		skeleton.show()
		return

	var frame : UltraleapFrame = tracking.get_last_frame()
	if frame == null:
		skeleton.hide()
		return
	if use_interpolation:
		var interpolated_frame : UltraleapFrame = tracking.get_interpolated_frame(Time.get_ticks_usec())
		if interpolated_frame != null:
			frame = interpolated_frame
	
	if handedness == UltraleapTypes.Chirality.Left:
		if frame.is_left_hand_visible:
			skeleton.show()
			set_hand(frame.left_hand)
		else:
			skeleton.hide()
			
	else:
		if frame.is_right_hand_visible:
			skeleton.show()
			set_hand(frame.right_hand)
		else:
			skeleton.hide()


func set_hand(hand : UltraleapHand):
	set_wrist(hand)

	set_finger(hand.thumb)
	set_finger(hand.index)
	set_finger(hand.middle)
	set_finger(hand.ring)
	set_finger(hand.pinky)
	
	set_palm(hand)
	
	
func set_palm(hand : UltraleapHand):
	var palm_name : String = ex_map.get("palm")
	
	var bone_idx : int = skeleton.find_bone(palm_name)
	
	if bone_idx == -1:
		return
	
	# Leaving it for now, we might find a need for a palm bone after all...
	#skeleton.set_bone_pose_rotation(bone_idx, hand.palm.orientation)
	#skeleton.set_bone_pose_position(bone_idx, hand.palm.position)
	#skeleton.set_bone_pose_scale(bone_idx, global_bone_scale)
	
	
func set_wrist(hand : UltraleapHand):
	var wrist_name : String = ex_map.get("wrist")
	
	var bone_idx : int = skeleton.find_bone(wrist_name)
	var wrist_transform : Transform3D = hand.wrist * tracking.rigging_transform
	
	skeleton.set_bone_global_pose_override(bone_idx, wrist_transform, 1.0)
	
func set_finger(finger : UltraleapDigit):
	if finger.type != UltraleapDigit.FingerType.Thumb:
		set_bone(finger, finger.metacarpal)
	set_bone(finger, finger.proximal)
	set_bone(finger, finger.intermediate)
	set_bone(finger, finger.distal)


func set_bone(finger : UltraleapDigit, bone : UltraleapBone):
	var bone_name : String = bone_name_from_type(bone.type)
	var finger_name : String = finger_name_from_type(finger.type)
	var bone_mapped_name : String = ex_map.get(finger_name + "_" + bone_name)
	if (bone == null or bone_mapped_name == ""):
		return
		
	set_bone_transform(ex_map.get(finger_name + "_" + bone_name), bone.get_transform())


func set_bone_transform(bone_name : String, bone_transform : Transform3D):
	var bone_idx : int = skeleton.find_bone(bone_name)
	if bone_idx == -1:
		print("Couldn't set bone " + bone_name)
		return
	skeleton.set_bone_pose_rotation(bone_idx, bone_transform.basis.get_rotation_quaternion())
	if rig_settings.set_joints_positions:
		skeleton.set_bone_pose_position(bone_idx, bone_transform.origin)


func finger_name_from_type(finger_type: UltraleapDigit.FingerType) -> String:
	return {
		UltraleapDigit.FingerType.Thumb: "thumb",
		UltraleapDigit.FingerType.Index: "index",
		UltraleapDigit.FingerType.Middle: "middle",
		UltraleapDigit.FingerType.Ring: "ring",
		UltraleapDigit.FingerType.Pinky: "pinky"
	}[finger_type] if finger_type != UltraleapDigit.FingerType.None else null


func bone_name_from_type(bone_type : UltraleapBone.BoneType) -> String:
	return {
		UltraleapBone.BoneType.Metacarpal: "metacarpal",
		UltraleapBone.BoneType.Proximal: "proximal",
		UltraleapBone.BoneType.Intermediate: "intermediate",
		UltraleapBone.BoneType.Distal: "distal",
		UltraleapBone.BoneType.Arm: "arm"
	}[bone_type] if bone_type != UltraleapBone.BoneType.None else null
